import numpy as np


def latlon2xyz(coords, R=1.0):
    """Convert geocentric coordaintes into 3D Cartesian."""
    lon = np.deg2rad(coords[:, 0])
    lat = np.deg2rad(coords[:, 1])

    x = R * np.cos(lat) * np.cos(lon)
    y = R * np.cos(lat) * np.sin(lon)
    z = R * np.sin(lat)

    return np.stack([x, y, z]).T


def latlon2stereographic(coords, R=1.0):
    """Stereographic projection centered on the North Pole."""
    lon = np.deg2rad(coords[:, 0])
    lat = np.deg2rad(coords[:, 1])

    k = 2 * R / (1 + np.sin(lat))

    x = k * np.cos(lat) * np.sin(lon)
    y = k * np.cos(lat) * np.cos(lon)

    return np.stack([x, y]).T



import numpy as np
import xarray as xr
from scipy.spatial import Delaunay

from . import transform


def compute_weights_delaunay(points, xi):
    """Compute remapping weights for linear interpolation.

    The interpolation is based on a Delaunay triangulation on the sphere [0].

    [0]: https://www.redblobgames.com/x/1842-delaunay-voronoi-sphere/

    Args:
        points (tuple[ndarrays]): Tuple with source grid coordinates.
        xi (tuple[ndarrays]): Tuple with target grid coordinates.

    Returns:
        xr.Dataset: Remapping information
            "src_idx": Array with source grid indices
            "weights": Remapping weights
            "valid": Boolean mask determining if a target point is within the
                triangulation. If it's outside, the target point will not
                receive any data after interpolation any user might want
                to set it to some missing value.

    See also:
        `apply_weights`
    """
    # Convert coordinates into stereographic and Cartesian coordinates
    # for triangulation and weight computation respectively.
    src_xy = transform.latlon2stereographic(np.array(points).T)
    src_xyz = transform.latlon2xyz(np.array(points).T)

    tgt_xy = transform.latlon2stereographic(np.stack(xi, axis=-1))
    tgt_xyz = transform.latlon2xyz(np.stack(xi, axis=-1))

    # Triangulation in stereographic projection
    tri = Delaunay(src_xy)
    triangles = tri.find_simplex(tgt_xy)
    src_idx = tri.simplices[triangles]
    valid = triangles >= 0

    # Compute barycentric weights in 3D
    verts_xyz = src_xyz[src_idx]
    v0, v1, v2 = verts_xyz[:, 0], verts_xyz[:, 1], verts_xyz[:, 2]
    p = tgt_xyz

    area_total = np.linalg.norm(np.cross(v1 - v0, v2 - v0), axis=1)
    w0 = np.linalg.norm(np.cross(v1 - p, v2 - p), axis=1) / area_total
    w1 = np.linalg.norm(np.cross(v2 - p, v0 - p), axis=1) / area_total
    w2 = 1.0 - w0 - w1

    weights = np.stack([w0, w1, w2], axis=-1)

    return xr.Dataset(
        data_vars={
            "src_idx": (("tgt_idx", "tri"), src_idx),
            "weights": (("tgt_idx", "tri"), weights),
            "valid": (("tgt_idx",), valid),
        }
    )


def apply_weights(var, src_idx, weights, valid):
    """Apply given remapping weights.

    Args:
        var (ndarray): Array to remap.
        kwargs: Remapping weights as returned by `compute_weights`.

    Returns:
        ndarray: Remapped values

    See also:
        `compute_weights`
    """
    return np.where(valid, (var[src_idx] * weights).sum(axis=-1), np.nan)



import warnings

import numpy as np
import cf_xarray as cf_xarray
import xarray as xr
import healpix

from ..resample import HEALPixResampler
from ..show import map_show, map_contour


def get_nest(dx):
    return dx.cf["grid_mapping"].healpix_order in ["nest", "nested"]


def get_nside(dx):
    try:
        return dx.cf["grid_mapping"].healpix_nside
    except (AttributeError, KeyError):
        if dx.squeeze().ndim > 1:
            raise ValueError(
                "Cannot infer the HEALPix resolution from a multidimensional dataset.\n"
                "Consider adding a coordinate reference system to the dataset or passing a one-dimensional array instead.\n"
                "See also: easygems.healpix.attach_coords\n"
                "Reference: https://easy.gems.dkrz.de/Processing/datasets/remapping.html#storing-the-coordinate-reference-system"
            )
        return healpix.npix2nside(dx.size)


def get_npix(dx):
    return healpix.nside2npix(get_nside(dx))


def get_extent_mask(dx, extent):
    lon = dx.lon
    lat = dx.lat

    w, e, s, n = extent  # Shortcut for N/S/E/W bounds

    is_in_lon = (lon - w) % 360 < (e - w) % 360  # consider sign change
    is_in_lat = (lat > s) & (lat < n)

    return is_in_lon & is_in_lat


def get_full_chunks(indices, chunksize):
    """Return indices of complete chunks, given a list of indices and a chunksize."""
    used_chunks = np.unique(np.asarray(indices) // chunksize)

    return (used_chunks[:, np.newaxis] * chunksize + np.arange(chunksize)).flatten()


def isel_extent(dx, extent):
    return np.arange(get_npix(dx))[get_extent_mask(dx, extent)]


def broadcast_array(dx, fill_value=np.nan):
    """Broadcast a limited-area HEALPix array to full shape."""
    arr = np.full_like(dx, fill_value, shape=get_npix(dx))
    arr[dx.cell] = dx.values

    return arr


def fix_crs(ds: xr.Dataset):
    # remove crs dimension (crs should really be 0-dimensional, but sometimes we keep a dimension
    # to be compatible with netcdf
    grid_mapping_var = ds.cf["grid_mapping"].name
    ds = ds.drop_vars(grid_mapping_var).assign_coords(
        {grid_mapping_var: ((), 0, ds.cf["grid_mapping"].attrs)}
    )
    return ds


def guess_crs(ds: xr.Dataset):
    warnings.warn(
        "No CRS coordinate was found. Attempting to infer it from the dataset shape. Please check the result!",
        stacklevel=4,
    )

    if "cell" in ds.dims:
        pix = ds.cell
    elif "values" in ds.dims:
        pix = ds.values
    elif "value" in ds.dims:
        pix = ds.value

    crs = xr.DataArray(
        name="crs",
        attrs={
            "grid_mapping_name": "healpix",
            "healpix_nside": healpix.npix2nside(pix.size),
            "healpix_order": "nest",
        },
    )
    return ds.assign_coords(crs=crs)


def attach_coords(ds: xr.Dataset, signed_lon=False):
    try:
        ds.cf["grid_mapping"]
    except KeyError:
        ds = guess_crs(ds)
    else:
        ds = fix_crs(ds)

    cell = ds.get("cell") if "cell" in ds.dims else np.arange(get_npix(ds))

    lons, lats = healpix.pix2ang(
        get_nside(ds), cell.astype("i8"), nest=get_nest(ds), lonlat=True
    )
    if signed_lon:
        lons = np.where(lons <= 180, lons, lons - 360)
    else:
        # Both healpy and healpix produce longitudes in the range [-45, 360]
        # While this is mathematically valid, it may be unexpected in Earth system science.
        lons %= 360
    return ds.assign_coords(
        cell=cell,
        lat=(
            ("cell",),
            lats,
            {"units": "degree_north", "standard_name": "latitude", "axis": "Y"},
        ),
        lon=(
            ("cell",),
            lons,
            {"units": "degree_east", "standard_name": "longitude", "axis": "X"},
        ),
    )


def healpix_show(var, method="nearest", nest=True, **kwargs):
    r = HEALPixResampler(nside=get_nside(var), nest=nest, method=method)

    return map_show(var, resampler=r, **kwargs)


def healpix_contour(var, method="nearest", nest=True, **kwargs):
    r = HEALPixResampler(nside=get_nside(var), nest=nest, method=method)

    return map_contour(var, resampler=r, **kwargs)


__all__ = [
    "get_nest",
    "get_nside",
    "get_npix",
    "get_extent_mask",
    "get_full_chunks",
    "isel_extent",
    "fix_crs",
    "attach_coords",
    "healpix_show",
    "healpix_contour",
]
